import logging
from enum import Enum
from typing import Union, List
from middleware.detectors.env_vars import EnvVarsDetector
from opentelemetry.resource.detector.azure.app_service import (
    AzureAppServiceResourceDetector,
)
from opentelemetry.resource.detector.azure.vm import AzureVMResourceDetector
from opentelemetry.resource.detector.azure.functions import (
    AzureFunctionsResourceDetector,
)
from opentelemetry.sdk.extension.aws.resource.ecs import AwsEcsResourceDetector
from opentelemetry.sdk.extension.aws.resource._lambda import AwsLambdaResourceDetector
from opentelemetry.sdk.extension.aws.resource.beanstalk import (
    AwsBeanstalkResourceDetector,
)
from opentelemetry.sdk.extension.aws.resource.ec2 import AwsEc2ResourceDetector
from opentelemetry.sdk.extension.aws.resource.eks import AwsEksResourceDetector
from opentelemetry.resourcedetector.gcp_resource_detector._detector import (
    GoogleCloudResourceDetector,
)

_logger = logging.getLogger(__name__)


# Define an enumeration
class Detector(Enum):
    ENVVARS = "envvars"
    AWS_LAMBDA = "aws_lambda"
    AWS_BEANSTALK = "aws_beanstalk"
    AWS_ECS = "aws_ecs"
    AWS_EKS = "aws_eks"
    AWS_EC2 = "aws_ec2"

    AZURE_VM = "azure_vm"
    AZURE_APP_SERVICE = "azure_app_service"
    AZURE_FUNCTIONS = "azure_functions"

    GCP = "gcp"


def process_detector_input(input_value: Union[str, List[Detector]]):
    """
    This function takes either a string of comma-separated detectors names
    or a list of Detector enum members and returns a list of Detector enum members.
    """
    unique_detectors = set()
    # If input is a string, split it into a list
    if isinstance(input_value, str):
        items = [item.strip() for item in input_value.split(",")]

        for item in items:
            try:
                # Attempt to access the enum member using the string
                enum_value = Detector[item.upper()]  # Convert to uppercase for matching
                unique_detectors.add(enum_value)
            except KeyError:
                _logger.warning(f"{item} is not a valid Detector")

    # If input is already a list, ensure it contains Detector enums
    elif isinstance(input_value, list):
        if all(isinstance(item, Detector) for item in input_value):
            unique_detectors.update(input_value)
        else:
            raise ValueError("All items in the list must be instances of Detector Enum")

    else:
        raise TypeError(
            "Input must be either a string or a list of Detector Enum members"
        )

    return list(unique_detectors)


def get_detectors(detectors: List[Detector]):
    resource_detectors = set()

    if detectors is not None and len(detectors) != 0:
        for detector in detectors:
            if detector == Detector.ENVVARS:
                resource_detectors.add(EnvVarsDetector())
            elif detector == Detector.AWS_LAMBDA:
                resource_detectors.add(AwsLambdaResourceDetector())
            elif detector == Detector.AWS_BEANSTALK:
                resource_detectors.add(AwsBeanstalkResourceDetector())
            elif detector == Detector.AWS_ECS:
                resource_detectors.add(AwsEcsResourceDetector())
            elif detector == Detector.AWS_EKS:
                resource_detectors.add(AwsEksResourceDetector())
            elif detector == Detector.AWS_EC2:
                resource_detectors.add(AwsEc2ResourceDetector())
            elif detector == Detector.AZURE_VM:
                resource_detectors.add(AzureVMResourceDetector())
            elif detector == Detector.AZURE_APP_SERVICE:
                resource_detectors.add(AzureAppServiceResourceDetector())
            elif detector == Detector.AZURE_FUNCTIONS:
                resource_detectors.add(AzureFunctionsResourceDetector())
            elif detector == Detector.GCP:
                resource_detectors.add(GoogleCloudResourceDetector())
    return list(resource_detectors)


# Example usage
detectors_string = "aws_lambda,gcp"
detectors_list = [Detector.AWS_LAMBDA, Detector.GCP]



import logging
import os
import requests
from opentelemetry.sdk.environment_variables import (
    OTEL_EXPORTER_OTLP_ENDPOINT,
    OTEL_LOG_LEVEL,
    OTEL_SERVICE_NAME,
)
from middleware.detectors.detector import Detector, process_detector_input
from typing import Union, List
from middleware.version import __version__
# Environment Variable Names
OTEL_SERVICE_VERSION = "OTEL_SERVICE_VERSION"
DEBUG = "DEBUG"
SAMPLE_RATE = "SAMPLE_RATE"

# Middleware Envs
MW_API_KEY = "MW_API_KEY"
MW_SERVICE_NAME = "MW_SERVICE_NAME"
MW_AGENT_SERVICE = "MW_AGENT_SERVICE"
MW_TARGET = "MW_TARGET"
MW_CUSTOM_RESOURCE_ATTRIBUTES = "MW_CUSTOM_RESOURCE_ATTRIBUTES"
MW_APM_COLLECT_TRACES = "MW_APM_COLLECT_TRACES"
MW_APM_COLLECT_METRICS = "MW_APM_COLLECT_METRICS"
MW_APM_COLLECT_LOGS = "MW_APM_COLLECT_LOGS"
MW_APM_COLLECT_PROFILING = "MW_APM_COLLECT_PROFILING"
MW_PROPAGATORS = "MW_PROPAGATORS"
MW_CONSOLE_EXPORTER = "MW_CONSOLE_EXPORTER"
MW_DEBUG_LOG_FILE = "MW_DEBUG_LOG_FILE"
MW_PROJECT_NAME = "MW_PROJECT_NAME"
MW_SAMPLE_RATE = "MW_SAMPLE_RATE"
MW_LOG_LEVEL = "MW_LOG_LEVEL"
MW_DETECTORS = "MW_DETECTORS"

OTEL_PROPAGATORS = "OTEL_PROPAGATORS"

# Default values
DEFAULT_TARGET = "http://localhost:9319"
DEFAULT_PORT = "9319"
DEFAULT_COLLECT_TRACES = True
DEFAULT_COLLECT_METRICS = True
DEFAULT_COLLECT_LOGS = True
DEFAULT_COLLECT_PROFILING = False
DEFAULT_AGENT_SERVICE = "localhost"
DEFAULT_EXPORTER_PROTOCOL = "grpc"
DEFAULT_SERVICE_NAME = f"python-service-{os.getpid()}"
DEFAULT_LOG_LEVEL = "INFO"
DEFAULT_PROPAGATORS = "b3"
DEFAULT_SAMPLE_RATE = 1


# DETECTORS
DETECT_ENVVARS = Detector.ENVVARS
DETECT_AWS_LAMBDA = Detector.AWS_LAMBDA
DETECT_AWS_BEANSTALK = Detector.AWS_BEANSTALK
DETECT_AWS_ECS = Detector.AWS_ECS
DETECT_AWS_EKS = Detector.AWS_EKS
DETECT_AWS_EC2 = Detector.AWS_EC2

DETECT_AZURE_VM = Detector.AZURE_VM
DETECT_AZURE_APP_SERVICE = Detector.AZURE_APP_SERVICE
DETECT_AZURE_FUNCTIONS = Detector.AZURE_FUNCTIONS

DETECT_GCP = Detector.GCP

# Errors and Warnings
MISSING_SERVICE_NAME_ERROR = (
    "Missing service name. Specify either "
    + "OTEL_SERVICE_NAME/MW_SERVICE_NAME environment variable or service_name in the "
    + "options parameter. If left unset, this will show up in middleware "
    + "as python-service-pid"
)

log_levels = {
    "NOTSET": logging.NOTSET,
    "DEBUG": logging.DEBUG,
    "INFO": logging.INFO,
    "WARN": logging.WARN,
    "WARNING": logging.WARNING,
    "ERROR": logging.ERROR,
    "FATAL": logging.FATAL,
    "CRITICAL": logging.CRITICAL,
}

_logger = logging.getLogger(__name__)


# pylint: disable=too-many-arguments,too-many-instance-attributes
class MWOptions:
    """
    Middleware Options used to configure the OpenTelemetry SDK.

    This class allows for the configuration of various middleware options for
    OpenTelemetry, with support for environment variables to override default
    values. If an option is set as an environment variable, it will take precedence
    over any parameter passed during instantiation. If neither is present, the
    default value will be used.

    **Attributes and Environment Variables:**

    - `access_token (str)`: API key for authentication.
      - Environment Variable: `MW_API_KEY`.
      - Example usage:
                access_token = "whkvkobudfitutobptgonaezuxpjjypnejbb"

    - `service_name (str)`: Name of the service.
      - Environment Variable: `MW_SERVICE_NAME` (alternative: `OTEL_SERVICE_NAME`).
      - Example usage:
                service_name = "MyPythonServer"

    - `collect_traces (bool)`: Flag to collect traces.
      - Environment Variable: `MW_APM_COLLECT_TRACES` (default: True).
      - Example usage:
                collect_traces = True

    - `collect_metrics (bool)`: Flag to collect metrics.
      - Environment Variable: `MW_APM_COLLECT_METRICS` (default: True).
      - Example usage:
                collect_metrics = False

    - `collect_logs (bool)`: Flag to collect logs.
      - Environment Variable: `MW_APM_COLLECT_LOGS` (default: True).
      - Example usage:
                collect_logs = True

    - `collect_profiling (bool)`: Flag to collect profiling.
      - Environment Variable: `MW_APM_COLLECT_PROFILING` (default: False).
      - Example usage:
                collect_profiling = True

    - `log_level (str)`: Logging level.
      - Environment Variable: `MW_LOG_LEVEL` (alternative: `OTEL_LOG_LEVEL`).
      - Example usage: (INFO, DEBUG, WARNING, ERROR)
                log_level = DEBUG

    - `mw_agent_service (str)`: Address of the middleware agent service.
      - Environment Variable: `MW_AGENT_SERVICE` (default: "localhost").
      - Use for containerized app with agent installed.
      - Example usage:
        (Docker)
                mw_agent_service = 172.17.0.1
        (Kubernetes)
                mw_agent_service = mw-service.mw-agent-ns.svc.cluster.local

    - `target (str)`: Target endpoint for the OTLP exporter.
      - Environment Variable: `MW_TARGET` (alternative: `OTEL_EXPORTER_OTLP_ENDPOINT`, default: "http://localhost:9319").
      - Use Target for Agent-less setup i.e agent is not installed.
      - Example usage:
                target = "https://myapp.middleware.io:443"

    - `custom_resource_attributes (str)`: Custom resource attributes.
      - Environment Variable: `MW_CUSTOM_RESOURCE_ATTRIBUTES`.
      - Example usage:
                custom_resource_attributes = "call_id=12345678, request_id=987654321"


    - `otel_propagators (str)`: Propagators for context propagation.
      - Environment Variable: `MW_PROPAGATORS` (alternative: `OTEL_PROPAGATORS`, default: "b3").
      - Example usage:(https://opentelemetry.io/docs/specs/otel/configuration/sdk-environment-variables/#:~:text=Known%20values%20for%20OTEL_PROPAGATORS%20are%3A)
                otel_propagators = "b3,tracecontext"

    - `console_exporter (bool)`: Flag to enable console exporter.It dumps telemetry data to console.
      - Environment Variable: `MW_CONSOLE_EXPORTER` (default: False).
      - Example usage:
                console_exporter = True

    - `debug_log_file (bool)`: Flag to enable debug logging to file. To dump telemetry data in files.
      - Environment Variable: `MW_DEBUG_LOG_FILE` (default: False).
      - Use only if console_exporter is used.
      - Example usage:
                debug_log_file = True

    - `project_name (str)`: Name of the project.
      - Environment Variable: `MW_PROJECT_NAME`.
      - Example usage:
                project_name = "TestingProject"

    - `sample_rate (int)`: Sample rate for telemetry data.
      - Environment Variable: `MW_SAMPLE_RATE`.
      - Should be the value from `0` to `1`.
        AlwaysOn (1), AlwaysOff (0), or a TraceIdRatio as 1/N
      - To use OTEL Sampler use ENVs and avoid this sample_rate.
      - Example usage:
                sample_rate = 0.5

    - `detectors (str,List[Detector])`: Detectors to add for aws, azure, gcp or envvars.
      - Environment Variable: `MW_DETECTORS`.
      - Example usage:
                detectors_string = "aws_lambda,gcp" for MW_DETECTORS
                detectors_list = [DETECT_AWS_LAMBDA, DETECT_GCP]

    **Defaults**:

    - Target: http://localhost:9319
    - Log Level: INFO
    - Collect Traces: True
    - Collect Metrics: True
    - Collect Logs: True
    - Agent Service: 'localhost'
    - Agent Service: 'localhost'
    - Service Name: 'unknown_service:python'
    - Propagators: 'b3'
    - Sample Rate: 1

    """

    access_token = None
    service_name = None
    collect_traces = DEFAULT_COLLECT_TRACES
    collect_metrics = DEFAULT_COLLECT_METRICS
    collect_logs = DEFAULT_COLLECT_LOGS
    collect_profiling = DEFAULT_COLLECT_PROFILING
    log_level = DEFAULT_LOG_LEVEL
    mw_agent_service = DEFAULT_AGENT_SERVICE
    target = DEFAULT_TARGET
    custom_resource_attributes = None
    otel_propagators = DEFAULT_PROPAGATORS
    console_exporter = False
    debug_log_file = False
    project_name = None
    sample_rate = None
    detectors = None

    def __init__(
        self,
        access_token: str = None,
        service_name: str = None,
        collect_traces: bool = DEFAULT_COLLECT_TRACES,
        collect_metrics: bool = DEFAULT_COLLECT_METRICS,
        collect_logs: bool = DEFAULT_COLLECT_LOGS,
        collect_profiling: bool = DEFAULT_COLLECT_PROFILING,
        log_level: str = DEFAULT_LOG_LEVEL,
        mw_agent_service: str = DEFAULT_AGENT_SERVICE,
        target: str = DEFAULT_TARGET,
        custom_resource_attributes: str = None,
        otel_propagators: str = DEFAULT_PROPAGATORS,
        console_exporter: bool = False,
        debug_log_file: bool = False,
        project_name: str = None,
        sample_rate: int = None,
        detectors: Union[str, List[Detector]] = None,
    ):
        self.access_token = os.environ.get(MW_API_KEY, access_token)
        self.service_name = os.environ.get(
            OTEL_SERVICE_NAME, os.environ.get(MW_SERVICE_NAME, service_name)
        )
        if not self.service_name:
            _logger.debug(MISSING_SERVICE_NAME_ERROR)
            self.service_name = DEFAULT_SERVICE_NAME

        self.collect_traces = parse_bool(MW_APM_COLLECT_TRACES, collect_traces)
        self.collect_metrics = parse_bool(MW_APM_COLLECT_METRICS, collect_metrics)
        self.collect_logs = parse_bool(MW_APM_COLLECT_LOGS, collect_logs)
        self.collect_profiling = parse_bool(MW_APM_COLLECT_PROFILING, collect_profiling)

        log_level = os.environ.get(
            OTEL_LOG_LEVEL, os.environ.get(MW_LOG_LEVEL, log_level)
        )
        if log_level and log_level.upper() in log_levels:
            self.log_level = log_level.upper()
            logging.basicConfig(level=log_levels[self.log_level])
        else:
            _logger.warning("invalid log_level:", log_level)

        self.target = os.environ.get(
            OTEL_EXPORTER_OTLP_ENDPOINT, os.environ.get(MW_TARGET, target)
        )

        if "https" not in self.target:
            self.mw_agent_service = os.environ.get(MW_AGENT_SERVICE, mw_agent_service)
            if self.mw_agent_service is not None:
                self.target = f"http://{self.mw_agent_service}:{DEFAULT_PORT}"

        self.custom_resource_attributes = os.environ.get(
            MW_CUSTOM_RESOURCE_ATTRIBUTES, custom_resource_attributes
        )

        self.otel_propagators = os.environ.get(
            OTEL_PROPAGATORS, os.environ.get(MW_PROPAGATORS, otel_propagators)
        )
        os.environ["OTEL_PROPAGATORS"] = self.otel_propagators    
        self.console_exporter = parse_bool(MW_CONSOLE_EXPORTER, console_exporter)
        self.debug_log_file = parse_bool(MW_DEBUG_LOG_FILE, debug_log_file)
        self.project_name = os.environ.get(MW_PROJECT_NAME, project_name)
        self.sample_rate = parse_int(MW_SAMPLE_RATE, sample_rate, DEFAULT_SAMPLE_RATE)
        self.detectors = os.environ.get(MW_DETECTORS, detectors)
        _health_check(options=self)
        _get_instrument_info(options=self)

def parse_bool(
    environment_variable: str, default_value: bool, error_message: str = None
) -> bool:
    """
    Attempts to parse the provided environment variable into a bool. If it
    does not exist or fails parse, the default value is returned instead.

    Args:
        environment_variable (str): the environment variable name to use
        default_value (bool): the default value if not found or unable parse
        error_message (str): the error message to log if unable to parse

    Returns:
        bool: either the parsed environment variable or default value
    """
    val = os.getenv(environment_variable, None)
    if val:
        try:
            if val.upper() == "FALSE":
                return False
            if val.upper() == "TRUE":
                return True
            return False
        except ValueError:
            if error_message is not None:
                _logger.warning(error_message)
    return default_value


def parse_int(
    environment_variable: str, param: int, default_value: int, error_message: str = None
) -> int:
    """
    Attempts to parse the provided environment variable into an int. If it
    does not exist or fails parse, the default value is returned instead.

    Args:
        environment_variable (str): the environment variable name to use
        param(int): fallback parameter to check before setting default
        default_value (int): the default value if not found or unable parse
        error_message (str): the error message to log if unable to parse

    Returns:
        int: either the parsed environment variable, param, or default value
    """
    val = os.getenv(environment_variable, None)
    if val:
        try:
            return int(val)
        except ValueError:
            if error_message is not None:
                _logger.warning(error_message)
            return default_value
    elif isinstance(param, int):
        return param
    else:
        return default_value

def _health_check(options: MWOptions):
    if options.target == "" or ("https" not in options.target) :
        try:
            response = requests.get(
                f"http://{options.mw_agent_service}:13133/healthcheck", timeout=5
            )
            if response.status_code != 200:
                _logger.warning(
                    "MW Agent Health Check is failing ...\nThis could be due to incorrect value of MW_AGENT_SERVICE\nIgnore the warning if you are using MW Agent older than 1.7.7 (You can confirm by running `mw-agent version`)"
                )
        except requests.exceptions.RequestException as e:
            _logger.warning(f"MW Agent Health Check is failing ...\nException while MW Agent Health Check:{e}")

def _get_instrument_info(options: MWOptions):
    _logger.debug(
        f"Middleware Python Instrumentation Started with SDK version: {__version__}"
    )
    if options.target == "" or ("https" not in options.target):
        _logger.debug(f"Using Agent Instrumentation")
    else:
        _logger.debug(f"Using Serverless Instrumentation")
    _logger.debug(f"OTLP Endpoint:{options.target}")
    if options.collect_metrics:
        _logger.debug(f"Metrics: Enabled")
    if options.collect_traces:
        _logger.debug(f"Traces: Enabled")
    if options.collect_logs:
        _logger.debug(f"Logs: Enabled")
    if options.collect_profiling:
        _logger.debug(f"Profiling: Enabled")



import logging
import inspect
import traceback
from typing import Optional, Type
import sys
from logging import getLogger
from typing import Optional
from opentelemetry.instrumentation.distro import BaseDistro
from middleware.metrics import create_meter_provider
from middleware.options import MWOptions, parse_bool
from middleware.resource import create_resource
from middleware.trace import create_tracer_provider
from middleware.log import create_logger_handler
from middleware.profiler import collect_profiling
from opentelemetry import trace
from opentelemetry.trace import Tracer, get_current_span, get_tracer, Span, get_tracer, Status, StatusCode
import os
import json

_logger = getLogger(__name__)

mw_tracker_called = False
distro_called = False
MW_TRACKER = "MW_TRACKER"
DEFAULT_TRACKER = False
isTracker = parse_bool(
    MW_TRACKER, DEFAULT_TRACKER, "Invalid MW_TRACKER: expected boolean"
)


def mw_tracker(
    options: Optional[MWOptions] = None,
):
    """
    Configures the OpenTelemetry SDK to send telemetry to middleware.

    Args:
        options (MWOptions, optional): the MWOptions used to
        configure the the SDK. These options can be set either as parameters
        to this function or through environment variables

    Example
    --------
    >>> from middleware import mw_tracker, MWOptions, record_exception, DETECT_AWS_EC2
    >>> mw_tracker(
    >>>     MWOptions(
    >>>         access_token="whkvkobudfitutobptgonaezuxpjjypnejbb",
    >>>         target="https://myapp.middleware.io:443",
    >>>         console_exporter=True,
    >>>         debug_log_file=True,
    >>>         service_name="MyPythonServer",
    >>>         otel_propagators = "b3,tracecontext",
    >>>         custom_resource_attributes="call_id=12345678, request_id=987654321",
    >>>         detectors=[DETECT_AWS_EC2]
    >>>     )
    >>> )

    """
    global mw_tracker_called, isTracker

    if not distro_called:
        return
    if not isTracker and mw_tracker_called:
        _logger.warning(
            "Skipping mw_tracker config: `export MW_TRACKER=True` to use config settings"
        )
        return
    if isTracker and mw_tracker_called:
        _logger.warning("Skipping mw_tracker config: overriding of config not allowed")
        return

    if options is None:
        options = MWOptions()

    _logger.debug(vars(options))
    resource = create_resource(options)
    if options.collect_traces:
        create_tracer_provider(options, resource)
    if options.collect_metrics:
        create_meter_provider(options, resource)
    if options.collect_logs:
        handler = create_logger_handler(options, resource)
        logging.getLogger().addHandler(handler)
    if options.collect_profiling:
        collect_profiling(options)    

    mw_tracker_called = True

def extract_function_code(tb_frame, lineno):
    """Extracts the full function body where the exception occurred."""
    try:
        source_lines, start_line = inspect.getsourcelines(tb_frame)
        end_line = start_line + len(source_lines) - 1
        
        if len(source_lines) > 20:
            # Get 10 lines above and 10 below the exception line
            start_idx = max(0, lineno - start_line - 10)
            end_idx = min(len(source_lines), lineno - start_line + 10)
            source_lines = source_lines[start_idx:end_idx]
        
        function_code = "".join(source_lines)  # Convert to a string
        
        return {
            "function_code": function_code,
            "function_start_line": start_line if len(source_lines) <= 20 else None,
            "function_end_line": end_line if len(source_lines) <= 20 else None,
        }    
        
    except Exception as e:
        return {
            "function_code": f"Error extracting function code: {e}",
            "function_start_line": None,
            "function_end_line": None
        }

# Replacement of span.record_exception to include function source code
def custom_record_exception(span: Span, exc: Exception):
    """Custom exception recording that captures function source code."""
    exc_type, exc_value, exc_tb = exc.__class__, str(exc), exc.__traceback__

    if exc_tb is None:
        span.set_attribute("exception.warning", "No traceback available")
        span.record_exception(exc)
        return

    tb_details = traceback.extract_tb(exc_tb)
    
    if not tb_details:
        span.set_attribute("exception.warning", "Traceback is empty")
        span.record_exception(exc)
        return

    stack_info = []
    
    for (frame, _), (filename, lineno, func_name, _) in zip(traceback.walk_tb(exc_tb), tb_details):
        function_details = extract_function_code(frame, lineno) if frame else "Function source not found."
  
        stack_entry = {
            "exception.file": filename,
            "exception.line": lineno,
            "exception.function_name": func_name,
            "exception.function_body": function_details["function_code"],
            "exception.start_line": function_details["function_start_line"],
            "exception.end_line": function_details["function_end_line"],
        }

        # Check if the file is from site-packages
        if "site-packages" in filename:
            stack_entry["exception.is_file_internal"] = "true"
        else:
            stack_entry["exception.is_file_internal"] = "false"

        stack_info.insert(0, stack_entry)  # Prepend instead of append

    # Determine if the exception is escaping
    current_exc = sys.exc_info()[1]  # Get the currently active exception
    exception_escaped = current_exc is exc  # True if it's still propagating

    mw_vcs_repository_url = os.getenv("MW_VCS_REPOSITORY_URL")
    mw_vcs_commit_sha = os.getenv("MW_VCS_COMMIT_SHA")
  
    # Serialize stack info as JSON string since OpenTelemetry only supports string values
    stack_info_str = json.dumps(stack_info, indent=2)
    
    # Add extra details in the existing "exception" event
    span.add_event(
        "exception",  
        {
            "exception.type": str(exc_type.__name__),
            "exception.message": exc_value,
            "exception.stacktrace": traceback.format_exc(),
            "exception.escaped": exception_escaped,
            "exception.vcs.commit_sha": mw_vcs_commit_sha or "",
            "exception.vcs.repository_url": mw_vcs_repository_url or "",
            "exception.stack_details": stack_info_str,  # Attach full stacktrace details
        }
    )

def record_exception(exc_type: Type[BaseException], exc_value: BaseException, exc_traceback) -> None:
    """
    Reports an exception as a span event, creating a dummy span if necessary.

    Args:
        exc_type (Type[BaseException]): The type of the exception.
        exc_value (BaseException): The exception instance.
        exc_traceback: The traceback object.

    Example
    --------
    >>> import sys
    >>> try:
    >>>     print("Divide by zero:", 1 / 0)
    >>> except Exception as e:
    >>>     sys.excepthook(*sys.exc_info())

    """
    # Retrieve the current span if available
    span = get_current_span()
    if span and span.is_recording():
        custom_record_exception(span, exc_value)
        return

    # Create a new span if none is found
    tracer: Tracer = get_tracer("mw-tracer")
    span_name = exc_type.__name__ if exc_type else "UnknownException"

    with tracer.start_span(span_name) as span:
        custom_record_exception(span, exc_value)
        span.set_status(Status(StatusCode.ERROR, str(exc_value)))


# pylint: disable=too-few-public-methods
class MiddlewareDistro(BaseDistro):
    """
    An extension of the base python OpenTelemetry distro, which provides
    a mechanism to automatically configure some of the more common options
    for users. This class is auto-detected by the `opentelemetry-instrument`
    command.

    This class doesn't need to be touched directly when using the distro. If
    you'd like to explicitly set configuration in code, use the
    mw_tracker() function above with `export MW_TRACKER=True`.

    [tool.poetry.plugins."opentelemetry_distro"]
    distro = "middleware.opentelemetry.distro:MiddlewareDistro"
    """

    def _configure(self, **kwargs):
        global isTracker, distro_called
        distro_called = True
        if not isTracker:
            mw_tracker()



import grpc
import sys
import logging
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import (
    BatchSpanProcessor,
    SimpleSpanProcessor,
    ConsoleSpanExporter,
)
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter
from opentelemetry.processor.baggage import ALLOW_ALL_BAGGAGE_KEYS, BaggageSpanProcessor
from middleware.options import MWOptions
from opentelemetry.trace import set_tracer_provider
from middleware.sampler import configure_sampler

_logger = logging.getLogger(__name__)


def create_tracer_provider(options: MWOptions, resource: Resource) -> TracerProvider:
    """
    Configures and returns a new TracerProvider to send traces telemetry.

    Args:
        options (MWOptions): the middleware options to configure with
        resource (Resource): the resource to use with the new tracer provider

    Returns:
        TracerProvider: the new tracer provider
    """
    exporter = OTLPSpanExporter(
        endpoint=options.target,
        compression=grpc.Compression.Gzip,
    )
    trace_provider = TracerProvider(
        resource=resource, shutdown_on_exit=True, sampler=configure_sampler(options)
    )
    trace_provider.add_span_processor(BaggageSpanProcessor(ALLOW_ALL_BAGGAGE_KEYS))
    trace_provider.add_span_processor(
        BatchSpanProcessor(
            exporter,
        )
    )
    if options.console_exporter:
        output = sys.stdout
        if options.debug_log_file:
            log_file = "mw-traces"
            try:
                output = open(log_file, "w")
            except Exception:
                _logger.error(f"Cannot open the log file for writing: {log_file}")
                output = sys.stdout
        trace_provider.add_span_processor(
            SimpleSpanProcessor(
                ConsoleSpanExporter(
                    out=output,
                )
            )
        )
    set_tracer_provider(tracer_provider=trace_provider)
    return trace_provider
